import requests

from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler


class CustomFacebookGraphReader:
    """
    Custom reader for the Facebook Graph API.
    """

    def __init__(self, creds_file, config_file=None, **kwargs):
        self._creds = load_file(creds_file, "yml")
        self._config = load_file(config_file, "yml")
        self.version = kwargs.get("version", "v11.0")

        self._request_session = requests.Session()
        self.paging = None
        self.data_set = []

    @retry_handler(
        exceptions=Exception, total_tries=5, should_raise=True, backoff_factor=5
    )
    def _graph_api_request_handler(self):
        """
        Basic handler for the Facebook api response.
        https://developers.facebook.com/docs/marketing-api/insights/parameters/v12.0
        API Parameters:
            action_attribution_windows:
                desc: The attribution window for the actions.
                default_value: default
                d_type: list(1d_view, 7d_view, 28d_view, 1d_click,
                             7d_click, 28d_click, dda, default)
            action_breakdowns:
                desc: How to break down action results.
                default_value: Vec
                d_type: list(action_device, action_canvas_component_name, action_carousel_card_id,
                            action_carousel_card_name, action_destination, action_reaction,
                            action_target_id, action_type, action_video_sound, action_video_type)
            action_report_time:
                desc: Determines the report time of action stats.
                default_value:
                d_type: list(impression, conversion, mixed)
            breakdowns:
                desc: How to break down the result.
                default_value:
                d_type: list(ad_format_asset, age, app_id, body_asset,
                            call_to_action_asset, country,
                            description_asset, gender, image_asset,
                            impression_device, link_url_asset,
                            product_id, region, skan_conversion_id,
                            title_asset, video_asset, dma,
                            frequency_value,
                            hourly_stats_aggregated_by_advertiser_time_zone,
                            hourly_stats_aggregated_by_audience_time_zone,
                            place_page_id, publisher_platform,
                            platform_position, device_platform)
            date_preset:
                desc: Represents a relative time range.
                default_value: last_30d
                d_type: list(today, yesterday, this_month, last_month,
                this_quarter, maximum,
                            last_3d, last_7d, last_14d, last_28d, last_30d,
                            last_90d, last_week_mon_sun,
                            last_week_sun_sat, last_quarter,
                            last_year, this_week_mon_today,
                            this_week_sun_today, this_year)
            default_summary:
                desc: Select fields on the exporting report file.
                default_value:
                d_type: list(bool)
            fields:
                desc: Fields to be retrieved.
                default_value: impressions and spend
                d_type: list(string)
            level:
                desc: Represents the level of result.
                default_value:
                d_type: str(ad, adset, campaign, account)
            time_range:
                desc: A single time range object. UNIX timestamp not supported.
                    This param is ignored if time_ranges is provided.
                default_value: dict('since':YYYY-MM-DD,'until':YYYY-MM-DD)
                d_type:
        """
        print("GET: graph.facebook.com.")
        try:
            response = self._request_session.get(
                # Note the since and until params does not seem to work at all,
                # so filtering after request
                url=f"https://graph.facebook.com/v12.0/{self._creds.get('act')}/insights",
                params={
                    "date_preset": self._config.get("date_preset", "last_30d"),
                    "time_increment": self._config.get("time_increment", 1),
                    "fields": self._config.get(
                        "fields", "impressions,spend,clicks,cpc,ctr,reach"
                    ),
                    "level": self._config.get("level", "adset"),
                    "limit": self._config.get("limit", 1000),
                    "access_token": self._creds["access_token"],
                },
            )

            print(f"Status: {response.status_code}.")
            if response.status_code != 200:
                raise EnvironmentError(f"Reason:{response.json()}.")

            # gather initial request, data and page token
            response_json = response.json()
            self.paging = response_json.get("paging", {}).get("next", None)
            print("Page Token: Initial Request.")
            self.data_set.append(response_json.get("data", []))

            # page through the response data
            while True:
                if self.paging is not None:
                    print(f"Page Token: {self.paging}.")
                    page_response = self._request_session.get(self.paging)
                    page_response_json = page_response.json()
                    if page_response.status_code != 200:
                        raise EnvironmentError(f"Reason:{page_response.json()}.")

                    self.paging = page_response_json.get("paging", None).get(
                        "next", None
                    )
                    self.data_set.append(page_response_json.get("data", []))
                else:
                    print("Paging data complete.")
                    break

            return self.data_set[0]

        except Exception as err:
            raise err

    def run_query(self):
        """
        Get metrics data from Facebook Graph API.
        The Pages API is a set of Facebook Graph API endpoints that apps can
        use to create and manage a Page's settings and content.
        Metric data of public Pages is stored by Facebook for 2 years.
        Metric data of unpublished Pages is stored for only 5 days.
        """
        response = self._graph_api_request_handler()
        if response is not None:
            return response
        else:
            print("Facebook returned no data for given period.")
            return None



def tracking_metadata(data_object: list, **kwargs) -> list:
    """
    A method to manage a more consistent tracking mechanism
    for any service that requires tracking to be sent to s3.

    Object consumed is expected to look like:
        "service",  # the name of the service i.e. google_analytics_v3
        "response_datetime",  # the datetime.now() of when the request was made
        "index",  # response index, page token etc
        "id_name",  # identifier name of the current request i.e. view_id
        "id_value",  # value of the identifier od the current request i.e 528447
        "config_datetime",  # the date scope that the config is querying
        "configuration", # dict object of the configuration file for reference
        "byte_size", # the size of the object in bytes
    """

    valid_keys = [
        "service",
        "response_datetime",
        "index",
        "id_name",
        "id_value",
        "config_datetime",
        "configuration",
        "byte_size",
    ]

    dataset = {}
    for key in valid_keys:
        try:
            dataset[key] = kwargs.get(key)
        except KeyError:
            raise KeyError(
                "The tracking_metadata only supports specific key value pairs."
            )

    data_object.append(dataset)
    return data_object



"""
    CUSTOM WRITER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
import gzip
import io
import json
import os

import boto3
from interface import implements

from sdc_dp_helpers.base_writer import BaseWriter, CustomLocalJsonWriter


class GAV3Writer:
    def __init__(self, bucket, folder_path, profile_name=None):
        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)
        self.s3_resource = self.boto3_session.resource("s3")
        self.bucket = bucket
        self.folder_path = folder_path

    def write_to_s3(self, payload):
        """
        This pulls the yielded dataset from the GA reader in a manner
        that consumes the dataset of the given view_id and date,
        and writes it to s3 so that duplication does not occur.
        :param payload: This is a key value object that looks like:
                        {
                            "data": list(),
                            "date": string,
                            "view_id": string
                        }
        """
        _view_id, _date = payload.get("view_id"), payload.get("date").replace("-", "")
        _data = payload.get("data")
        write_path = f"{self.folder_path}/{_view_id}/{_date}.json"
        print(
            f"Writing data to s3://{self.bucket}/{write_path} partitioned by view_id and date."
        )
        self.s3_resource.Object(self.bucket, write_path).put(Body=json.dumps(_data))


class CustomS3GZJsonWriter(CustomLocalJsonWriter, implements(BaseWriter)):
    """Class Extends Basic LocalGZJsonWriter"""

    def __init__(
        self,
        file_name: str,
        folder_path: str,
        bucket: str,
        profile_name: str = None,
        **kwargs,
    ):
        self.os_path_sep = "/"

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)

        self.bucket = bucket
        """Writes a general object to s3"""
        self.s3_resource = self.boto3_session.resource("s3")
        super().__init__(file_name=file_name, folder_path=folder_path, **kwargs)
        self.metadata_file = kwargs.get("metadata_file", "metadata.json.gz")

    def set_full_path(self, nested_dirs: list = None):
        """Set full path to write to"""
        # creates full path
        full_path = [self.folder_path] + nested_dirs
        return os.path.join(*full_path).replace("\\", "/")

    # pylint: disable=no-member,too-many-arguments
    def write_to_s3(
        self,
        json_data,
        data_path,
        mode="w",
        encoding="utf-8",
        ensure_ascii=False,
        default=None,
    ):
        """
        Write report to s3 bucket
        """
        buffer = io.BytesIO()
        with gzip.GzipFile(fileobj=buffer, mode=mode) as file_:
            with io.TextIOWrapper(file_, encoding=encoding) as textio_obj:
                textio_obj.write(
                    json.dumps(json_data, ensure_ascii=ensure_ascii, default=default)
                )
        # write to s3
        buffer.seek(0)
        self.s3_resource.Bucket(self.bucket).put_object(
            Key=os.path.join(self.full_path, data_path).replace("\\", "/"), Body=buffer
        )

    def write_metadata(self, metadata: dict):
        """ "
        Creates required GA/ViewID/metricName structure
        """
        # create directory structure if not exists
        self.full_path = self.set_full_path(
            nested_dirs=[metadata.get("viewID", "."), metadata.get("metricName", "")]
        )
        self.write_to_s3(json_data=metadata, data_path=self.metadata_file, mode="w")

    def write_file(self):
        """
        upload python dict into s3 bucket with gzip archive
        """
        # generate a unix timestamp suffix for file
        suff = self.timestamp_suffix()
        data_path = "{}_{}.json.gz".format(self.file_name, suff)
        self.write_to_s3(self.data, data_path, mode="wb")
        self.data = []



import os, ast, json, requests

import datetime
from xero import Xero
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import request_handler
from sdc_dp_helpers.xero import auth_handler as pixie
from sdc_dp_helpers.xero.config_managers import get_config
from sdc_dp_helpers.xero.writers import CustomS3JsonWriter


class CustomXeroReader:
    """
    Custom Xero Reader
    """

    def __init__(self, **kwargs):
        self.config_path = kwargs.get("config_path")
        self.creds_path = kwargs.get("creds_path")

        self.config = get_config(self.config_path)
        
        today = datetime.date.today()
        first = today.replace(day=1)
        last_month = first - datetime.timedelta(days=1)

        self.config["from_date"] = self.config.get( "from_date", last_month.strftime("%Y-%m-01") )

        self.config["to_date"] = self.config.get( "to_date", today.strftime("%Y-%m-%d") )

        self.client_id = self.config.get("client_id", None)
        if not self.client_id:
            raise ValueError("No client_id set")

        self.refresh_token_bucket = self.config.get("refresh_token_bucket")
        self.refresh_token_path = self.config.get("refresh_token_path")
        self.auth_token = pixie.get_auth_token(
            self.client_id,
            {
                "token_path": self.refresh_token_path,
                "bucket_name": self.refresh_token_bucket,
                "local_token_path": self.creds_path,
            },
            None,  # "default" #"ritdu-sdc-eu1"
        )

        self.writer = CustomS3JsonWriter()

    # def update_refresh_token(self):
    #     """
    #     Write the refresh token to s3 to be used in the next Auth run.
    #     """
    #     self.writer.write_to_s3(
    #         data=self.auth_token,
    #         filename=f"{self.client_id}_auth_token",
    #         suffix="json",
    #     )

    def traverse_reports(self) -> list:
        """
        Loops through reports in the config to pull each of them
        """
        for report_name in self.config.get("reports", []):
            if report_name not in [
                "BalanceSheet",
                "ProfitAndLoss",
                # "TrialBalance", #not pulling this atm
                "AgedPayablesByContact",
                "AgedReceivablesByContact",
            ]:
                raise ValueError(report_name + " is not supported or does not exist.")

            self.xero_obj = Xero( self.auth_token )
            if report_name in ["BalanceSheet", "ProfitAndLoss"]:
                trackingcategories = (
                    i for i in self.xero_obj.trackingcategories.all() if i != None
                )

                request_params = {
                    "report_name": report_name,
                    "tenant_id": tenant_id,
                }

                if report_name == "BalanceSheet":
                    request_params.update(
                        {
                            "from_date": self.config.get("from_date"),
                            "to_date": self.config.get("to_date"),
                        }
                    )
                elif report_name == "ProfitAndLoss":
                    request_params.update({"date": self.config.get("from_date")})

                data_set = []
                for tenant_id in self.config.get("tenant_ids", []):
                    data_set += self.fetch_report(request_params)

            elif report_name in ["AgedPayables", "AgedReceivables"]:
                request_params = {
                    "report_name": report_name,
                    "tenant_id": tenant_id,
                }

                request_params.update({"date": self.config.get("from_date")})

                data_set = []
                # the below request filters all contacts for this tenant to contacts with the account_number
                for contact_id in self.xero_obj.contacts.filter(
                    AccountNumber__startswith="999999"
                ):
                    request_params.update({"contactId": contact_id})
                    data_set += self.fetch_report(request_params)

            self.writer.write_to_s3(
                data_set,
                {
                    "data_variant": "reports",
                    "collection_name": report_name,
                    "tenant_name": data_set.get("Reports").get("ReportTitles")[1],
                    "date": datetime.datetime.strptime(
                        data_set.report.get("Reports")
                        .get("ReportTitles")[3]
                        .split(" to ")[0]
                        .strip(),
                        "%d %B %Y",
                    ).strftime("%Y-%m-%d"),
                },
            )

        return data_set

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 0.1)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    def fetch_report(self, request: dict) -> str:
        """
        This method accepts Parameters
            report_name: the report name as per Xero API endpoint
            from_date: the start_date of the report
            to_date: the end date of the report
        and returns the report for those parameters using the requests module to access the API directly
        """
        # unpack request
        report_name = request["report_name"]
        from_date = request.get( "from_date", request.get("date", None) )
        to_date = request.get( "to_date", datetime.date.today().strftime('%Y-%m-%d') )
        tracking_category = request["tracking_category"]
        filter_items = (
            None if "filter_items" not in request.keys() else request["filter_items"]
        )
        self.auth_token = pixie.get_auth_token(
            self.client_id,
            {
                "token_path": self.refresh_token_path,
                "bucket_name": self.refresh_token_bucket,
                "local_token_path": self.creds_path,
            },
            None,  # "default" #"ritdu-sdc-eu1"
        )
        # self.auth_token.tenant_id = request['tenant_id']

        my_headers = {
            "Authorization": "Bearer " + self.auth_token.token["access_token"], # self.xero_client.token["access_token"],
            "Xero-Tenant-Id": request['tenant_id'],
            "Accept": "application/json",
        }
        my_params = {
            "fromDate": from_date,
            "toDate": to_date,
        }
        if tracking_category:
            my_params.update({ "trackingCategoryID": tracking_category["TrackingCategoryID"] } )

        my_params.update(filter_items)
        response = requests.get(
            "https://api.xero.com/api.xro/2.0/Reports/" + report_name,
            params=my_params,
            headers=my_headers,
        )
        # response.text = response.text.strip("'<>() ").replace('\'', '\"')
        # print(response.text)
        report = json.loads(
            response.text.replace("\r", "").replace("\n", "").strip("'<>() ")
        )
        report = {
            "tenantId": self.auth_token.tenant_id,
            "trackingCategoryId": tracking_category["TrackingCategoryID"],
            "report": report,
            "from_date": from_date,
            "to_date": to_date,
        }
        return json.dumps(report)

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 0.1)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    def run_request(self, xero_client, api_object, request):
        """
        Run the API request that consumes a request payload and site url.
        This separates the request with the request handler from the rest of the logic.
        """
        # ToDo: Handle API Errors
        api_call = getattr(xero_client, api_object)
        return api_call.filter(
            from_date=request.get("from_date", None),
            to_date=request.get("to_date", None),
            page=request.get("page", None),
        )

    def fetch_modules(self):
        """
        Consumes a .yaml config file and loops through the date and url
        to return relevant data from Xero API.
        """
        self.auth_token = pixie.refresh_auth_token(self.client_id, self.auth_token, self.creds_path )
        xero = Xero(self.auth_token)

        from_date = self.config.get("from_date", None)
        to_date = self.config.get("to_date", None)
        data_set = []

        for api_object in self.config.get("modules", []):
            if api_object not in [
                "contacts",
                "accounts",
                "invoices",
                "banktransactions",
                "manualjournals",
                "purchaseorders",
            ]:
                raise ValueError(api_object + " is not supported or does not exist.")
            prev_response = None
            page = 1

            while True:
                response = self.run_request(
                    xero_client=xero,
                    api_object=api_object,
                    request={"from_date": from_date, "to_date": to_date, "page": page},
                )
                if len(response) < 1:
                    break
                elif response == prev_response:
                    break
                else:
                    data_set.append(json.dumps(ast.literal_eval(response)))

                # ensure the token is still fresh
                self.auth_token = pixie.refresh_auth_token(self.client_id, self.auth_token, self.creds_path)

                prev_response = response
                page += 1

    def run_main(self):
        """
        As dictated by the config;
        This function fetches all modules and reports requested based on the config
        """
        if getattr(self.config, "modules", None):
            self.fetch_modules()

        if getattr(self.config, "reports", None):
            self.traverse_reports()



import json

import yaml


def load_file(file_location, fmt):
    """
    Gathers file data from json or yaml.
    """
    if file_location is not None:
        if fmt in ["yaml", "yml"]:
            with open(file_location, "r") as file:
                config = yaml.safe_load(file)
        elif fmt == "json":
            with open(file_location, "r") as file:
                config = json.load(file)

    return config



import requests

from xero.auth import OAuth2Credentials

from xero.exceptions import *

from oauth2 import *
import boto3
import ast


def get_auth_token(client_id: str, token_details: dict, profile_name=None):
    """
    Consumes the client id and the previous auth processes refresh token.
    This returns an authentication token that will last 30 minutes
    to make queries the minute it is used. Or it will expire in 60 days of no use.
    The newly generated last refresh token now needs token stored for
    next use.
    PS: we receive and save the auth_token in a local dir supplied by the encapsulating project
    we never interact with s3 from inside of this class
    """
    if not client_id:
        raise ValueError("Invalid client_id")
    if len(client_id) != 32:
        raise ValueError("Invalid client_id")

    local_token_path = token_details["local_token_path"]
    auth_token = None
    with open( local_token_path, "r" ) as token_file:
        auth_token = ast.literal_eval( token_file.read() )
    print("original token:", auth_token )
    auth_creds = OAuth2Credentials( client_id, client_secret="", token=auth_token )

    return refresh_auth_token( client_id, auth_creds, local_token_path )


def refresh_auth_token( client_id: str, auth_creds: OAuth2Credentials, token_path: str ) -> OAuth2Credentials:
    """
    A simple handle of the token expiration.
    """
    if not auth_creds.expired():
        return auth_creds

    cred = {
        "grant_type": "refresh_token",
        "refresh_token": auth_creds.token["refresh_token"],  #
        "client_id": client_id,
    }
    response = requests.post("https://identity.xero.com/connect/token", cred)
    
    print("response", response)
    auth_token = response.json()
    print("auth token: ", auth_token)
    err_message = auth_token.get("error", None)
    if err_message:
        raise Exception(err_message)

    print("refreshed token:", auth_token)
    # save_auth_token(auth_token)
    print("Writing new Xero token to path: ", token_path)
    with open(token_path, 'w') as outfile:
        outfile.write( json.dumps(auth_token) )

    return OAuth2Credentials(client_id, client_secret="", token=auth_token)

def save_auth_token(auth_token):
    """
    function to persist the latest auth token to s3
    """
    # TODO: pull update_refresh_token from readers into here
    
    pass



from datetime import datetime, timedelta

import pandas as pd


def date_handler(date_string, date_format="%Y-%m-%d"):
    """
    Takes a date string or phrase and returns the valid date in
    the specified format.
    :date_string: A phrase like "3_days_ago", "yesterday" or "today"
                     or a date string such as "2021-01-01".
    :date_format: The returned formatting of the date.
    """
    try:
        if date_string == "today":
            return datetime.now().strftime(date_format)
        if date_string == "yesterday":
            return (datetime.now() - timedelta(days=1)).strftime(date_format)
        if "_days_ago" in date_string:
            return phrase_to_date(phrase=date_string, date_format=date_format)
        return date_string
    except ValueError as err:
        raise ValueError(
            "StartDate requires a valid input "
            'such as "today", "yesterday" or "<int>_days_ago".'
        ) from err


def phrase_to_date(phrase, date_format="%Y-%m-%d"):
    """
    Takes a typical phrase such as "3_days_ago" and returns a
    date based on that phrase.
    :phrase: str. Something like 3_days_ago or 25_days_ago etc.
    :return: '%Y-%m-%d'
    """
    try:
        date_delta = int(phrase.split("_")[0])
        return (datetime.now() - timedelta(days=date_delta)).strftime(date_format)
    except ValueError as err:
        raise ValueError(
            f"Phrasing for date: {phrase} is not valid, "
            f"try something like: 3_days_ago"
        ) from err


def date_range(start_date, end_date, delta=timedelta(days=1)):
    """
    The range is inclusive, so both start_date and end_date will be returned.
    :start_date: The datetime object representing the first day in the range.
    :end_date: The datetime object representing the second day in the range.
    :delta: A datetime.timedelta instance, specifying the step interval. Defaults to one day.
    Yields:
        Each datetime object in the range.
    """

    start_date = date_handler(date_string=start_date)
    end_date = date_handler(date_string=end_date)

    start_date = datetime.strptime(start_date, "%Y-%m-%d")
    end_date = datetime.strptime(end_date, "%Y-%m-%d")

    current_date = start_date
    while current_date <= end_date:
        yield current_date.strftime("%Y-%m-%d")
        current_date += delta


def filter_data_by_dates(
    start_date, end_date, date_field, data_frame: pd.DataFrame, date_fmt="%Y-%m-%d"
):
    """
    Takes a data frame and filters the data by given date field and
    date scopes that can be phrases.
    If no scope is added, imply current time is expected.
    """

    _now = datetime.now().strftime(date_fmt)
    if start_date is None:
        sd = _now
    else:
        sd = date_handler(start_date, date_fmt)

    if end_date is None:
        ed = _now
    else:
        ed = date_handler(end_date, date_fmt)

    print(f"Gathering data between {sd} and {ed}.")
    # filter data by given date field
    if sd is not None and ed is not None:
        data_frame["tmp_date"] = pd.to_datetime(data_frame[date_field])
        data_frame["tmp_date"] = data_frame["tmp_date"].dt.strftime(date_fmt)
        data_frame = data_frame[
            (
                data_frame["tmp_date"]
                >= datetime.strptime(sd, date_fmt).strftime(date_fmt)
            )
            & (
                data_frame["tmp_date"]
                <= datetime.strptime(ed, date_fmt).strftime(date_fmt)
            )
        ]
        data_frame = data_frame.drop("tmp_date", axis="columns")

    if len(data_frame.index) > 0:
        return data_frame

    print(f"No data for given date filter: {sd} to {ed}.")
    return None



"""
    utils functions
"""
import yaml


def init_config(file_name: str = "./example.yaml"):
    """
    Util method that creates an example config at the given file_name
    args:
        file_name (str): full path to config YAML file
    """
    config = {
        "dateRanges": [{"startDate": "2020-02-01", "endDate": "2020-02-01"}],
        "samplingLevel": "LARGE",
        "pageSize": 100,
        "chunk_date_range": True,
        "dayFreq": 1,
        "viewIDs": ["<add client viewID here>"],
        "queryConfig": [
            {
                "metrics": [{"expression": "ga:sessions", "alias": "sessions"}],
                "dimensions": [
                    {"name": "ga:date"},
                    {"name": "ga:medium"},
                    {"name": "ga:source"},
                    {"name": "ga:keyword"},
                    {"name": "ga:hostname"},
                    {"name": "ga:deviceCategory"},
                ],
                "orderBys": [{"fieldName": "ga:date", "sortOrder": "DESCENDING"}],
            }
        ],
    }

    # write to file
    with open(file_name, "w") as file_:
        yaml.dump(config, file_)
